"""
Droplet Deformation (DropDeform)

Determine droplet deformation into Polyhedron Shape.

Source code of DropDeform: https://git.rwth-aachen.de/avt-fvt/public/droplet-deformation/

The source code of DropDeform is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2025 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev

The coyprights and license terms are given in LICENSE.

This script contains all functions related to Bond number correlation.
Script version date: 06.01.2025
""" 

""" Import packages """
import sys
import numpy as np
import pandas as pd
from scipy.optimize import minimize
""" Functions """
def calc_pars_bo(config, poly, M_deform, shape_names, shape_names_ind):
    """ Function to determine fitting parameters of the Bond correlation """
    # Case differentation: fitting method
    # 0 = fit each polyhedron shape individual
    # 1 = fit general parameters to all polyhedron shapes
    # Prediefine dictionary to store the fitting parameter for the r_a =f(Bo) and r_f = f(Bo) correlations: 
    # D = r_a[D_rows x D_columns], r_f[D_rows x D_columns]
    # D_rows: contains the fitting parameter par1 and par2 and the MAPE of the fit as headers
    # D_columns: contains polyhedron shapes as headers
    # Define the parameters and polyhedron shapes
    D_rows = ['par1', 'par2', 'mape']
    # Fit to individual shapes
    if config.fit_meth == 0:
        # Create two nested dictionaries with zeros
        D_ra = {param: {shape: 0 for shape in shape_names} for param in D_rows}
        D_rf = {param: {shape: 0 for shape in shape_names} for param in D_rows}
        D_eps = {param: {shape: 0 for shape in shape_names} for param in D_rows}
        # Combine the two dictionaries into one
        D_bo = {
            'r_a': D_ra,
            'r_f': D_rf,
            'eps': D_eps
        }
        # Parameter bounds for r_a and r_f
        bnds_bo = {
            'r_a': [(config.zero, 5.0), (config.zero, 1.0)],  # Bounds for r_a parameters par1 and par2
            'r_f': [(config.zero, 2.0), (config.zero, 6.0)],   # Bounds for r_f parameters par1 and par2
            'eps': [(config.zero, 1.0), (config.zero, 2.0)]   # Bounds for eps parameters par1 and par2
        }
        # Initial guess values for r_a and r_f parameters
        initial_bo = {
            'r_a': [3.5, 0.7],  # Initial guess for r_f parameters
            'r_f': [0.8, 4.0],   # Initial guess for r_a parameters
            'eps': [0.5, 1.5]   # Initial guess for eps parameters
        }
        # Solver method
        solver_method_bo = 'SLSQP'
        # Iterate through the five polyhedron shapes
        for ish in range(5):
            #
            sys.stdout.write(f'Fitting Bond number correlation parameters for {shape_names_ind[ish]}...')
            # Polyhedron shape specific constants:
            # Polyhedron volume
            C_V = poly.C_V[ish]
            # Polyhedron insphere radius 
            C_rs = poly.C_rs[ish] 
            # Polygon face area
            C_A = poly.C_A[ish]
            # Sample data for r_a fit
            x_sampled, y_sampled = sample_data(config, M_deform, 0, ish)
            # Perform the parameter fitting for r_a = f(Bo) correlation
            result_ra = minimize(
                solver_bo_ind,                 # Objective function
                initial_bo['r_a'],          # Initial guess for parameters
                args=(x_sampled, y_sampled, 0, None, None, None),    # Additional arguments
                method=solver_method_bo,    # Optimization method (e.g., 'Powell')
                bounds=bnds_bo['r_a'],      # Bounds for the parameters
                tol=1e-10                   # Tolerance for convergence
            )
            # Obtain the values of the minimized function
            ra_mape = result_ra.fun
            # Extract result 
            par_ra1, par_ra2 = result_ra.x
            # Sample data for r_f fit
            x_sampled, y_sampled = sample_data(config, M_deform, 1, ish)
            # Perform the parameter fitting for r_f = f(Bo) correlation
            result_rf = minimize(
                solver_bo_ind,                 # Objective function
                initial_bo['r_f'],          # Initial guess for parameters
                args=(x_sampled, y_sampled, 1, C_V, C_A, None),    # Additional arguments
                method=solver_method_bo,    # Optimization method (e.g., 'Powell')
                bounds=bnds_bo['r_f'],      # Bounds for the parameters
                tol=1e-10                   # Tolerance for convergence
            )
            # Obtain the values of the minimized function
            rf_mape = result_rf.fun
            # Extract result 
            par_rf1, par_rf2 = result_rf.x
            # Sample data for a fit
            x_sampled, y_sampled = sample_data(config, M_deform, 2, ish)
            # Perform the parameter fitting for eps = f(Bo) correlation
            result_eps = minimize(
                solver_bo_ind,                 # Objective function
                initial_bo['eps'],          # Initial guess for parameters
                args=(x_sampled, y_sampled, 2, C_V, None, C_rs),    # Additional arguments
                method=solver_method_bo,    # Optimization method (e.g., 'Powell')
                bounds=bnds_bo['eps'],      # Bounds for the parameters
                tol=1e-10                   # Tolerance for convergence
            )
            # Obtain the values of the minimized function
            eps_mape = result_eps.fun
            # Extract result 
            par_eps1, par_eps2 = result_eps.x
            # Assign results and errors to dictionary D_bo
            D_bo['r_a']['par1'][shape_names_ind[ish]] = par_ra1
            D_bo['r_a']['par2'][shape_names_ind[ish]] = par_ra2
            D_bo['r_a']['mape'][shape_names_ind[ish]] = ra_mape
            D_bo['r_f']['par1'][shape_names_ind[ish]] = par_rf1
            D_bo['r_f']['par2'][shape_names_ind[ish]] = par_rf2
            D_bo['r_f']['mape'][shape_names_ind[ish]] = rf_mape
            D_bo['eps']['par1'][shape_names_ind[ish]] = par_eps1
            D_bo['eps']['par2'][shape_names_ind[ish]] = par_eps2
            D_bo['eps']['mape'][shape_names_ind[ish]] = eps_mape
            #
            sys.stdout.write("completed\n")

        return D_bo
    # Fit to all considered shapes
    if config.fit_meth == 1:
        # Predefine list
        x_sampled_set_ra = np.zeros((config.n_steps**3, 5))
        y_sampled_set_ra = np.zeros((config.n_steps**3, 5))
        x_sampled_set_rf = np.zeros((config.n_steps**3, 5))
        y_sampled_set_rf = np.zeros((config.n_steps**3, 5))
        x_sampled_set_eps = np.zeros((config.n_steps**3, 5))
        y_sampled_set_eps = np.zeros((config.n_steps**3, 5))        
        # Iterate through the five polyhedron shapes
        for ish in range(5):
            # Sample data for ra fit
            x_sampled_ra, y_sampled_ra = sample_data(config, M_deform, 0, ish)
            # Sample data for rf fit
            x_sampled_rf, y_sampled_rf = sample_data(config, M_deform, 1, ish,)
            # Sample data for eps fit
            x_sampled_eps, y_sampled_eps = sample_data(config, M_deform, 2, ish)
            # Determine the length of the array dynamically
            n_entries_ra = len(x_sampled_ra)
            # Assign data to list
            x_sampled_set_ra[0:n_entries_ra,ish] = x_sampled_ra
            y_sampled_set_ra[0:n_entries_ra,ish] = y_sampled_ra
            # Determine the length of the array dynamically
            n_entries_rf = len(x_sampled_rf)
            # Assign data to list
            x_sampled_set_rf[0:n_entries_rf,ish] = x_sampled_rf
            y_sampled_set_rf[0:n_entries_rf,ish] = y_sampled_rf
            # Determine the length of the array dynamically
            n_entries_eps = len(x_sampled_eps)
            # Assign data to list
            x_sampled_set_eps[0:n_entries_eps,ish] = x_sampled_eps
            y_sampled_set_eps[0:n_entries_eps,ish] = y_sampled_eps
        #
        sys.stdout.write('Fitting Bond number correlation parameters for...')
        # Create three dictionaries for D_ra, D_rf and D_eps, each with parameters initialized to zero
        D_ra = {param: 0 for param in D_rows}
        D_rf = {param: 0 for param in D_rows}
        D_eps = {param: 0 for param in D_rows}
        # Combine the three dictionaries into one
        D_bo = {
            'r_a': D_ra,
            'r_f': D_rf,
            'eps': D_eps
        }        
        # Parameter bounds for r_a, r_f and eps
        bnds_bo = {
            'r_a': [(config.zero, 2.0), (config.zero, 1.0)],  # Bounds for r_a parameters par1 and par2
            'r_f': [(config.zero, 5.0), (config.zero, 10.0)],   # Bounds for r_f parameters par1 and par2
            'eps': [(-10.0, config.zero), (config.zero, 3.0), (config.zero, 2.0)]   # Bounds for eps parameters par1 and par2
        }
        # Initial guess values for r_a, r_f and eps parameters
        initial_bo = {
            'r_a': [1.3, 0.7],  # Initial guess for r_a parameters
            'r_f': [2.5, 5.5],   # Initial guess for r_f parameters
            'eps': [-3.5, 1.6, 1.5]   # Initial guess for eps parameters
        }
        # Solver method
        solver_method_bo = 'SLSQP'
        # Perform the parameter fitting for r_a = f(Bo) correlation
        result_ra = minimize(
                solver_bo_gen,              # Objective function
                initial_bo['r_a'],          # Initial guess for parameters
                args=(poly, x_sampled_set_ra, y_sampled_set_ra, 0),         # Additional arguments
                method=solver_method_bo,    # Optimization method (e.g., 'Powell')
                bounds=bnds_bo['r_a'],      # Bounds for the parameters
                tol=1e-10                   # Tolerance for convergence
            )
        # Obtain the values of the minimized function
        ra_mape = result_ra.fun
        # Extract result 
        par_ra1, par_ra2 = result_ra.x
        # Perform the parameter fitting for r_f = f(Bo) correlation
        result_rf = minimize(
            solver_bo_gen,                 # Objective function
            initial_bo['r_f'],          # Initial guess for parameters
            args=(poly, x_sampled_set_rf, y_sampled_set_rf, 1),    # Additional arguments
            method=solver_method_bo,    # Optimization method (e.g., 'Powell')
            bounds=bnds_bo['r_f'],      # Bounds for the parameters
            tol=1e-10                   # Tolerance for convergence
        )
        # Obtain the values of the minimized function
        rf_mape = result_rf.fun
        # Extract result 
        par_rf1, par_rf2 = result_rf.x
        # Perform the parameter fitting for r_f = f(Bo) correlation
        result_eps = minimize(
            solver_bo_gen,                 # Objective function
            initial_bo['eps'],          # Initial guess for parameters
            args=(poly, x_sampled_set_eps, y_sampled_set_eps, 2),    # Additional arguments
            method=solver_method_bo,    # Optimization method (e.g., 'Powell')
            bounds=bnds_bo['eps'],      # Bounds for the parameters
            tol=1e-10                   # Tolerance for convergence
        )
        # Obtain the values of the minimized function
        eps_mape = result_eps.fun
        # Extract result 
        par_eps1, par_eps2, par_eps3 = result_eps.x
        # Assign results and errors to dictionary D_bo
        D_bo['r_a']['par1'] = par_ra1
        D_bo['r_a']['par2'] = par_ra2
        D_bo['r_a']['mape'] = ra_mape
        D_bo['r_f']['par1'] = par_rf1
        D_bo['r_f']['par2'] = par_rf2
        D_bo['r_f']['mape'] = rf_mape
        D_bo['eps']['par1'] = par_eps1
        D_bo['eps']['par2'] = par_eps2
        D_bo['eps']['par3'] = par_eps3
        D_bo['eps']['mape'] = eps_mape
        #
        sys.stdout.write("completed\n")

        return D_bo

def sample_data(config, M_deform, cor_bo, ish):
    """Function to sample the data for the Bo number fit.
    First data is transformedinto logarithmic scale, then sampled, then back transformed. """
    # Extract ra data from matrix
    # Case differentiation: Bo number correlation.
    # cor_bo = 0 --> r_a, cor_bo = 1 --> r_f
    if cor_bo == 0: 
        data_y = M_deform[0,:,1+ish]
    elif cor_bo == 1: 
        data_y = M_deform[1,:,1+ish]
    elif cor_bo == 2:
        data_y = M_deform[2,:,1+ish]
    # Delete zero values
    data_y_new = data_y[data_y != 0]
    # Extract Bo number data from matrix.
    data_x = M_deform[0,:,0]
    # Delete zero values
    data_x_new = data_x[data_y != 0]
    # Transform Bond numbers to logarithmic scale, distribute them to bins
    # choose an equal number of randomly chosen iterative solution results within these bins?    
    if config.log_scale == True:
        # Transfer into logarithmic scale
        data_x_new = np.log(data_x_new)
        # Combine x_data and y_data into a DataFrame for easy manipulation
        data = pd.DataFrame({'x': data_x_new, 'y': data_y_new})
        # Bin x into equally spaced bins
        data['x_bin'] = pd.cut(data['x'], bins=config.n_bins, labels=False)
        # Calculate the minimum number of points in any bin
        min_points_in_bin = data.groupby('x_bin').size().min()  
        # Set samples_per_bin based on the minimum points in any bin
        if min_points_in_bin > 0:
            # Limit to 5 or available points
            samples_per_bin = min(5, min_points_in_bin)  
        else:
            # No sampling possible if there are no points in any bin
            samples_per_bin = 0  
        # Sample from each bin
        balanced_sample = (data.groupby('x_bin', group_keys=False)
                        .apply(lambda group: group.sample(min(len(group), samples_per_bin), random_state=0)[['x', 'y']], include_groups=False)
                        .reset_index(drop=True))
        # Define x samples for fitting
        x_sampled = balanced_sample['x'].values
        # Transfer back into non logarithmic scale
        x_sampled = np.exp(x_sampled)
        # Define y samples for fitting
        y_sampled = balanced_sample['y'].values 
    else:
        x_sampled = data_x_new
        y_sampled = data_y_new

    return x_sampled, y_sampled

def solver_bo_ind(fpars_bo, x_sampled, y_sampled, cor_bo, C_V, C_A, C_rs):
    """ Solver to fit the Bond number correlation parameters for each polyhedron shape individualy """
    # Assign optimizing parameters
    par1, par2 = fpars_bo
    # Case differentiation: Bo number correlation.
    # cor_bo = 0 --> r_a, cor_bo = 1 --> r_f 
    if cor_bo == 0:     
        #print(y_sampled.shape)
        # r_a = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
        # Left hand side of equation 
        LHS = y_sampled
        # Right hand side of equation 
        #RHS = (1.0-(1.0-par1/(x_sampled+par1))**0.5) # original
        RHS = (1.0-(1.0-par1/((x_sampled**par2)+par1))**0.5)
    elif cor_bo == 1: 
        # r_f = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
        # Left hand side of equation 
        LHS = y_sampled
        # Parameter
        C_1 = ((C_A/np.pi)**0.5) * ((4.0*np.pi/(3.0*C_V))**(1.0/3.0))
        # Right hand side of equation 
        RHS = C_1*((1.0-par2/(x_sampled+par2))**0.5)
    elif cor_bo == 2: 
        # a = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
        # Left hand side of equation 
        LHS = y_sampled
        # Parameter
        C_6 = (4.0*np.pi/3.0) * (C_rs**3.0) / C_V
        # Right hand side of equation 
        RHS = C_6 + (1.0-C_6)*((1.0-par2/(x_sampled+par2))**0.5) 
    # Avoid division by zero. Except the zero values, which are outliers
    non_zero_LHS = LHS != 0
    # MAPE 
    mape = 100*np.mean((np.abs(LHS[non_zero_LHS] - RHS[non_zero_LHS]) / LHS[non_zero_LHS]))

    return mape

def solver_bo_gen(fpars_bo, poly, x_sampled_set, y_sampled_set, cor_bo):
    """ Solver to fit the general Bond number correlation parameters for all polyhedron shapes """
    # Assign optimizing parameters
    #par1, par2 = fpars_bo
    # Predefne lists
    LHS = []
    RHS = []
    # Iterate through polyhedron shapes
    for ish in range(5): 
        # Data for fit
        x_sampled = x_sampled_set[:,ish]
        y_sampled = y_sampled_set[:,ish]
        # Delete zero values
        x_sampled = x_sampled[x_sampled != 0]
        y_sampled = y_sampled[y_sampled != 0] 
        # Polyhedral shape parameters:
        # Number of polygon faces
        n_f = poly.n_f[ish]
        # Number of edges 
        n_e = poly.n_e[ish]
        # Case differentiation: Bo number correlation.
        # cor_bo = 0 --> r_a, cor_bo = 1 --> r_f 
        if cor_bo == 0:
            par1, par2 = fpars_bo
            # Parameter relation 1
            par1_rel = par1 * (n_e**0.3)
            par2_rel = par2 * (n_e**0.1)
            # r_a = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
            # Left hand side of equation 1
            LHS_set = y_sampled
            # Right hand side of equation 1
            RHS_set = (1.0-(1.0-par1_rel/((x_sampled**par2_rel)+par1_rel))**0.5)
        elif cor_bo == 1:
            par1, par2 = fpars_bo
            # Parameter relation 1
            par1_rel = par1 / (n_f**0.6) 
            # Parameter relation 2
            par2_rel = par2 / (n_e**0.1)
            # r_f = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
            # Left hand side of equation 1
            LHS_set = y_sampled
            # Right hand side of equation 1
            RHS_set = par1_rel*((1.0-par2_rel/(x_sampled+par2_rel))**0.5)
        elif cor_bo == 2:
            par1, par2, par3 = fpars_bo
            # Parameter relation 1
            par1_rel = par1 + (1.0-par1)*((1.0-par2/(n_f+par2))**0.5)
            # Parameter relation 2
            par2_rel = par3 / (n_e**0.1)
            # r_f = f(Bo) correlation (Henschke 1995, Ch. 4.3.2, Eq. 4.54)
            # Left hand side of equation 1
            LHS_set = y_sampled
            # Right hand side of equation 1
            RHS_set = par1_rel + (1-par1_rel)*((1.0-par2_rel/(x_sampled+par2_rel))**0.5)
        # Add data to list
        LHS.append(LHS_set)
        RHS.append(RHS_set)
    # Concatenate lists to ensure a single row for each
    LHS = np.hstack(LHS)
    RHS = np.hstack(RHS)
    # Avoid division by zero. Except the zero values, which are outliers
    non_zero_LHS = LHS != 0
    # MAPE 
    mape = 100*np.mean((np.abs(LHS[non_zero_LHS] - RHS[non_zero_LHS]) / LHS[non_zero_LHS]))

    return mape