"""
Droplet Deformation (DropDeform)

Determine droplet deformation into Polyhedron Shape.

Source code of DropDeform: https://git.rwth-aachen.de/avt-fvt/public/droplet-deformation/

The source code of DropDeform is licensed under the Eclipse Public License v2.0 (EPL 2.0).
Copyright (c) 2025 Fluid Process Engineering (AVT.FVT), RWTH Aachen University.
Written by Stepan Sibirtsev

The coyprights and license terms are given in LICENSE.

This script contains all functions related to droplet deformation model (three equation system and its solution).
Script version date: 06.01.2025
""" 

""" Import packages """
import numpy as np
from scipy.optimize import minimize
from tqdm import tqdm
""" Functions """
def calc_deform(config, poly, p, sigma, d, M_deform, shape_names_ind):
    """ Function to calculate droplet deformation by following steps:
    - calculate hydrostatic pressure acting on droplet
    - determine contact number
    - determine deformation shape (polyhedron shape)
    - determine deformation parameter: polyhedron edge lenght a, droplet's curvature radius r_a, droplet's contact area radius r_f
    """
    # Error tracking: deformation algorithm
    err_deform = 0
    # Iterate through the five polyhedron shapes
    for ish in range(5):
        print('Deformation simulation for:')
        # Use tqdm to track progress
        with tqdm(total=config.n_steps**3, desc=(shape_names_ind[ish]), unit='particle') as pbar:
            # Predefine counter
            k = 0
            # Iterate through pressure steps
            for ip in range(config.n_steps):
                # Considered pressure
                p_i = p[ip]
                # Iterate through interfacial tensions
                for ig in range(config.n_steps): 
                    # Considered interfacial tension
                    sigma_i = sigma[ig]
                    # Iterate through droplet diameter steps
                    for id in range(config.n_steps): 
                        # Considered droplet diameter
                        d_i = d[id]
                        # Droplet volume
                        V_i = calc_drop_vol(d_i)
                        # Polyhedron shape specific constants:
                        # Polyhedron volume
                        C_V = poly.C_V[ish]
                        # Polygon face area
                        C_A = poly.C_A[ish]
                        # Polygon incircle radius
                        C_rc = poly.C_rc[ish]
                        # Polyhedron midsphere radius
                        C_rm = poly.C_rm[ish]
                        # Polyhedron insphere radius 
                        C_rs = poly.C_rs[ish] 
                        # Number of edges 
                        n_e = poly.n_e[ish]
                        """ Calculate deformation based on the polyhedron shape """
                        # Lower bound of polyhedron edge length. 
                        # Resulting polyhedron volume must not be less than the droplet volume 
                        a_min = (V_i / C_V)**(1.0/3.0) 
                        # Upper bound of polyhedron edge length. 
                        # Polyhedron volume must not be greater than when the droplet touches its polygon face
                        a_max = (d_i / 2.0) / C_rs 
                        # Lower bound of droplet's curvature radius.
                        # Condition at hydrostatic pressure acting on drolet equal to infinite: r_a --> 0
                        # Near zero but not equalt to zero to avoid numerical issues.
                        r_amin = 1e-9
                        # Upper bound of droplet's curvature radius.
                        # Condition at hydrostatic pressure acting on drolet equal to zero: r_a --> r_i
                        r_amax = d_i / 2.0
                        # Lower bound of droplet's contact area radius.
                        # Condition at hydrostatic pressure acting on drolet equal to zero: r_f --> 0
                        # Near zero but not equalt to zero to avoid numerical issues.
                        r_fmin = 1e-9
                        # Upper bound of droplet's contact area radius.
                        # Condition at hydrostatic pressure acting on drolet equal to infinite: r_f --> polygon incircle radius r_s
                        # Difficult to set value since dependence on polyhedron edge lenght, thus d_i 
                        r_fmax = a_min*(C_A/np.pi)**(0.5)
                        # Bounds
                        bnds_deform = [(a_min, a_max), (r_amin, r_amax), (r_fmin, r_fmax)]
                        # Initial guess values for polyhedron edge length
                        a_0 = a_min + (a_max - a_min) / 2.0
                        # Initial guess values for droplet's curvature radius
                        r_a0 = (r_amax - r_amin) / 2.0
                        # Initial guess values for droplet's contact area radius
                        r_f0 = (C_rc * a_0) / 2.0
                        # Initial guess values
                        initial_deform = [a_0, r_a0, r_f0]
                        # Calculate Bond number
                        # Solver method
                        Bo = p_i * d_i / sigma_i
                        if Bo <= 1.0:
                            a_solver_method = 'Nelder-Mead'
                        else:
                            a_solver_method = 'Powell'
                        # Perform the optimization
                        result = minimize(solver_deform, 
                                        initial_deform, 
                                        args=(p_i, sigma_i, d_i, C_V, C_A, C_rc, C_rm, C_rs, n_e), 
                                        method=a_solver_method, 
                                        bounds=bnds_deform, 
                                        tol=1e-10)
                        # Obtain the values of the minimized function
                        solver_deform_err = result.fun
                        if solver_deform_err > 1.0:
                            err_deform = err_deform + 1
                            k = k+1
                            # Update the progress bar
                            pbar.update(1)
                            # dont include values with MAPE > 5% --> filtering outliers
                            continue
                        # Extract result 
                        a, r_a, r_f = result.x
                        # Polyhedron volume
                        V_p = C_V * (a**3.0)
                        # Local hold-up
                        eps = V_i / V_p
                        # Assign deformation parameters to M_deform
                        M_deform[0,k,0] = Bo
                        M_deform[1,k,0] = Bo
                        M_deform[2,k,0] = Bo
                        M_deform[0,k,ish+1] = 2*r_a / d_i
                        M_deform[1,k,ish+1] = 2*r_f / d_i
                        M_deform[2,k,ish+1] = eps
                        # Update counter
                        k = k+1
                        # Update the progress bar
                        pbar.update(1)

    # Print error deformation
    print('Results with MAPE > 1%:', err_deform)

    return M_deform

def solver_deform(vars, p_i, sigma_i, d_i, C_V, C_A, C_rc, C_rm, C_rs, n_e):
    """ Solver to calculate the polyhedron edge length 
    Variables:
    - polyhedron edge length a
    - droplet's curvature radius r_a
    - droplet's contact area radius r_f """
    # Assign optimizing parameters
    a, r_a, r_f = vars
    # Equation 1 (pressure relation): A_f * p_f = A_p * p_i
    # Contact area
    A_f = np.pi * r_f**2.0
    # Polygon face area
    A_p = C_A * a**2.0
    # Droplet's contact pressure
    p_f = (2.0 * sigma_i) / (r_a * (2.0 - r_a/(d_i/2.0)))
    # Relation between pressures:
    # Left hand side of equation 1
    LHS1 = A_f * p_f
    # Right hand side of equation 1
    RHS1 = A_p * p_i
    # MAPE for equation 1
    mape1 = 100*(np.abs(LHS1 - RHS1) / LHS1)
    # Equation 2 (geometric relation): r_o + x = r_c
    # Polygon incircle radius
    r_c = C_rc * a
    # Angle between polygon incircle radius r_c and midsphere radius r_m in radians
    alpha = np.arccos(C_rc / C_rm)
    # Distance between droplet's contact area radius and polygon edge
    x = r_a / np.tan(alpha)
    # Incircle radius of the polygon representing the contact area A_f
    r_o = ((np.pi*(C_rc**2.0)/C_A)**(0.5)) * r_f
    # Left hand side of equation 2
    LHS2 = r_o + x
    # Right hand side of equation 2
    RHS2 = r_c
    # MAPE for equation 2 
    mape2 = 100*(np.abs(LHS2 - RHS2) / LHS2)
    # Equation 3 (volume relation): V_i = V_p - n_e * V_c
    # Droplet Volume
    V_i = calc_drop_vol(d_i)
    # Polyhedron volume
    V_p = C_V*(a**3.0) 
    # Area of an isosceles triangl
    A_t = (x**2.0)*np.sin(2.0*alpha)/2.0
    # Opposite Angle of alpha. Delta is located in the same right-angled triangle as alpha
    delta = np.pi/2.0 - alpha
    # Area of circle segment
    A_a = (r_a**2.0)*(2.0*delta-np.sin(2.0*delta))/2.0
    # Channel area
    A_c = A_t - A_a
    # Correction factor
    C_cor = 2.0*(C_V-4.0*np.pi*(C_rs**3.0)/3.0)/(
        n_e*(C_rs**2.0)*((np.sin(2.0*alpha)/((np.tan(alpha))**2.0))-(np.pi-2.0*alpha-np.sin(2.0*alpha))))
    # Channel volume
    V_c = C_cor*a*A_c
    # Left hand side of equation 3
    RHS3 = V_i
    # Right hand side of equation 3
    LHS3 = V_p - n_e*V_c
    # MAPE for equation 3 
    mape3 = 100*(np.abs(LHS3 - RHS3) / LHS3)
    # Total error
    total_error = (mape1 + mape2 + mape3) / 3.0

    return total_error

def calc_drop_vol(d):
    """ Function to determine a droplet volume """
    # Sphere volume
    V_s = (np.pi/6.0)*(d**3.0)

    return V_s